#include <stdio.h>

#include "mex.h"
#include "cuda_runtime.h"
#include "..\GPUgaussMLEv2\GPUgaussMLEv2.cu"

void CUDAERRROR(const char *instr) {
    cudaError_t errornum;
    const char *str;
    if (errornum=cudaGetLastError()) {
        str=cudaGetErrorString(errornum);
        printf("%s\n", str);
        printf("You should clear this function in MATLAB for proper operation.\n", str);
    }
}

void mexFunction(int nlhs, mxArray *plhs[],	int	nrhs, const	mxArray	*prhs[]) {
    int blockx;
    int threadx;
    const mwSize *datasize;
    float PSFSigma, Ax, Ay, Bx, By, gamma, d, PSFSigma_y;
    int iterations, sz, Nfits;
    float *data, *d_data;
    float *d_Parameters,*d_CRLBs,*d_LogLikelihood;
    int Ndim, Nfitraw, fittype;

    //input checks
    if (nrhs<3)
        mexErrMsgTxt("Data input must include: data, PSFSigma, iterations, fittype !\n");
    
    if (mxGetClassID(prhs[0])!=mxSINGLE_CLASS)
        mexErrMsgTxt("Data must be comprised of single floats!\n");
    
    datasize=mxGetDimensions(prhs[0]);
    Ndim=mxGetNumberOfDimensions(prhs[0]);
    
    if (Ndim==2)Nfitraw=1;else Nfitraw=datasize[2];
    
    if (datasize[0] > IMSZBIG)
        mexErrMsgTxt("X,Y dimension of data must be smaller than 21.\n");
    if (datasize[1]!=datasize[0])
        mexErrMsgTxt("Fit Box must be square");
    sz=datasize[0];
    // mexPrintf("c sizeX: %d Ndims: %d\n",sz,Ndim);
    
    //put as many images as fit into a block
    //float szsqaure = sz*sz;
    int BlockSize = floor((float)15000/4/sz/sz);
    BlockSize = max(4, BlockSize);
    BlockSize = min(BSZ, BlockSize);
    
    //mexPrintf("cdasdas %d\n",BlockSize);
    
    Nfits=BlockSize*ceil( (float) Nfitraw/BlockSize);
    Nfits=Nfitraw;
    //mexPrintf("c sizeX: %d Nfits: %d\n", sz, Nfits);
    //mexPrintf("Nfits: %d, Nfitraw: %d\n", Nfits, Nfitraw);
    
    //get variables
    data=(float *) mxGetData(prhs[0]);
    PSFSigma=(float)mxGetScalar(prhs[1]); //matlab-dip_image convention
    iterations=(int) mxGetScalar(prhs[2]);
    fittype = (int) mxGetScalar(prhs[3]);
    
    if (nrhs>4)
        Ax = (float) mxGetScalar(prhs[4]);
    if (nrhs>5)
        Ay = (float) mxGetScalar(prhs[5]);
    if (nrhs>6)
        Bx = (float) mxGetScalar(prhs[6]);
    if (nrhs>7)
        By = (float) mxGetScalar(prhs[7]);
    if (nrhs>8)
        gamma = (float) mxGetScalar(prhs[8]);
    if (nrhs>9)
        d = (float) mxGetScalar(prhs[9]);
    if (nrhs>10)
        PSFSigma_y = (float) mxGetScalar(prhs[10]);
    else
        PSFSigma_y =PSFSigma;
		
	printf("sz: %d, Nfitraw: %d, PSFSigma: %f, iterations: %d, fittype: %d\n", sz, Nfitraw, PSFSigma, iterations, fittype);
/*	for (int i=0;i<sz*Nfitraw;i++) {
		printf("%f ", data[i]);
	}
	printf("\n");
*/	
    //create device variable for data and copy to device
    cudaMalloc((void**)&d_data, sz*sz*Nfits*sizeof(float));
    cudaMemset(d_data, 0, sz*sz*Nfits*sizeof(float));
    cudaMemcpy(d_data, data, sz*sz*Nfitraw*sizeof(float), cudaMemcpyHostToDevice);
    
    //create output for parameters and CRLBs
    switch(fittype){
        case 1: // (x,y,bg,I)
            cudaMalloc((void**)&d_Parameters,   NV_P*Nfits*sizeof(float));
            cudaMalloc((void**)&d_CRLBs,        NV_P*Nfits*sizeof(float));
            plhs[0]=mxCreateNumericMatrix(Nfitraw, NV_P, mxSINGLE_CLASS, mxREAL);
            plhs[1]=mxCreateNumericMatrix(Nfitraw, NV_P, mxSINGLE_CLASS, mxREAL);
            break;
        case 2: // (x,y,bg,I,Sigma)
        case 3: // (x,y,bg,I,z)
            cudaMalloc((void**)&d_Parameters,   NV_PS*Nfits*sizeof(float));
            cudaMalloc((void**)&d_CRLBs,        NV_PS*Nfits*sizeof(float));
            plhs[0]=mxCreateNumericMatrix(Nfitraw, NV_PS, mxSINGLE_CLASS, mxREAL);
            plhs[1]=mxCreateNumericMatrix(Nfitraw, NV_PS, mxSINGLE_CLASS, mxREAL);
            break;      
        case 4: // (x,y,bg,I,Sx,Sy)
            cudaMalloc((void**)&d_Parameters,   NV_PS2*Nfits*sizeof(float));
            cudaMalloc((void**)&d_CRLBs,        NV_PS2*Nfits*sizeof(float));
            plhs[0]=mxCreateNumericMatrix(Nfitraw, NV_PS2, mxSINGLE_CLASS, mxREAL);
            plhs[1]=mxCreateNumericMatrix(Nfitraw, NV_PS2, mxSINGLE_CLASS, mxREAL);
            break;
    }
    cudaMalloc((void**)&d_LogLikelihood,        Nfits*sizeof(float));
    plhs[2]=mxCreateNumericMatrix(Nfitraw, 1, mxSINGLE_CLASS, mxREAL);
    
    //setup kernel
    blockx = ceil( (float)Nfits/(float)BlockSize);
    threadx= BlockSize;
    
    dim3 dimBlock(threadx);
    dim3 dimGrid(blockx);
    
    printf("threadx: %d,blockx: %d,Nfitraw: %d\n", threadx, blockx, Nfitraw);
    
    switch(fittype) {
        case 1: //fit x,y,bg,I
            kernel_MLEFit<<<dimGrid, dimBlock>>>(d_data, PSFSigma, sz, iterations, d_Parameters, d_CRLBs, d_LogLikelihood, Nfits);
            break;
            
        case 2: //fit x,y,bg,I,sigma
            kernel_MLEFit_sigma<<<dimGrid, dimBlock>>>(d_data, PSFSigma, sz, iterations, d_Parameters, d_CRLBs, d_LogLikelihood, Nfits);
            break;
            
        case 3: //fit x,y,bg,I,z
            kernel_MLEFit_z<<<dimGrid, dimBlock>>>(d_data, PSFSigma, Ax, Ay, Bx, By, gamma, d, PSFSigma_y, sz, iterations, d_Parameters, d_CRLBs, d_LogLikelihood, Nfits);
            //mexPrintf("A:%f B:%f gamma:%f d:%f \n",Ax,Bx,gamma,d);
            break;
            
        case 4: //fit x,y,bg,I,sigmax,sigmay
            kernel_MLEFit_sigmaxy<<<dimGrid, dimBlock>>>(d_data, PSFSigma, sz, iterations, d_Parameters, d_CRLBs, d_LogLikelihood, Nfits);
            
    }
    CUDAERRROR("kernel");
    
    //copy to matlab output
/*	float * Parameters;
	Parameters = new float [Nfitraw*NV_PS];*/
    switch(fittype){
        case 1: // (x,y,bg,I)
            cudaMemcpy((float *)mxGetData(plhs[0]), d_Parameters, NV_P*Nfitraw*sizeof(float), cudaMemcpyDeviceToHost);
            cudaMemcpy((float *)mxGetData(plhs[1]), d_CRLBs, NV_P*Nfitraw*sizeof(float), cudaMemcpyDeviceToHost);
            break;
        case 2: // (x,y,bg,I,Sigma)
        case 3: // (x,y,bg,I,z)
            cudaMemcpy((float *)mxGetData(plhs[0]), d_Parameters, NV_PS*Nfitraw*sizeof(float), cudaMemcpyDeviceToHost);
 //           cudaMemcpy((float *)Parameters, d_Parameters, NV_PS*Nfitraw*sizeof(float), cudaMemcpyDeviceToHost);
            cudaMemcpy((float *)mxGetData(plhs[1]), d_CRLBs, NV_PS*Nfitraw*sizeof(float), cudaMemcpyDeviceToHost);
            break;
        case 4: // (x,y,bg,I,Sx,Sy)
            cudaMemcpy((float *)mxGetData(plhs[0]), d_Parameters, NV_PS2*Nfitraw*sizeof(float), cudaMemcpyDeviceToHost);
            cudaMemcpy((float *)mxGetData(plhs[1]), d_CRLBs, NV_PS2*Nfitraw*sizeof(float), cudaMemcpyDeviceToHost);
            break;
    }
    cudaMemcpy((float *)mxGetData(plhs[2]), d_LogLikelihood, Nfitraw*sizeof(float), cudaMemcpyDeviceToHost);

/*	for (int i=0; i<NV_PS*Nfitraw; i++) {
		printf("%f ", Parameters[i]);
	}
	printf("\n");
*/    
    //cleanup
    cudaFree(d_Parameters);
    cudaFree(d_CRLBs);
    cudaFree(d_LogLikelihood);
    return;
}
